(* http://www.cs.uofs.edu/~mccloske/courses/se507/alg_specs_lec.html *)

Module MyStack.

Require Import List.

Definition stack (A:Type) := list A.

Definition empty {A:Type} : stack A := nil.

Definition is_empty {A:Type} (s : stack A) : bool :=
  match s with
  | nil => true
  | _::_ => false
  end.

Definition push {A:Type} (x : A) (s : stack A) : stack A :=
  x::s.

Definition peek {A:Type} (s : stack A) : option A :=
  match s with
  | nil => None 
  | x::_ => Some x
  end.

Definition pop {A:Type} (s : stack A) : option (stack A) :=
  match s with
  | nil => None 
  | _::xs => Some xs
  end.

Definition size {A:Type} (s : stack A) : nat := 
  length s.

(*
is_empty empty = true
is_empty (push _ _) = false
peek empty = None
peek (push x _) = Some x
pop empty = None
pop (push _ s) = Some s
size empty = 0
size (push _ s) = 1 + size s
*)

Theorem empty_is_empty : forall (A:Type), 
  @is_empty A empty = true.
Proof. auto. Qed.

Theorem push_not_empty : forall (A:Type) (x:A) (s : stack A),
  is_empty(push x s) = false.
Proof. auto. Qed. 

Theorem peek_empty : forall (A:Type),
  @peek A empty = None.
Proof. auto. Qed.

Theorem peek_push : forall (A:Type) (x:A) (s : stack A),
  peek(push x s) = Some x.
Proof. auto. Qed.

Theorem pop_empty : forall (A:Type),
  @pop A empty = None.
Proof. auto. Qed.

Theorem pop_push : forall (A:Type) (x:A) (s : stack A),
  pop(push x s) = Some s.
Proof. auto. Qed.

Theorem size_empty : forall (A:Type),
  @size A empty = 0.
Proof. auto. Qed.

Theorem size_push : forall (A:Type) (x:A) (s : stack A),
  size(push x s) = 1 + size s.
Proof. auto. Qed.

End MyStack.


Extract Inductive bool => "bool" [ "true" "false" ].
Extract Inductive option => "option" [ "Some" "None" ].
Extract Inductive list => "list" [ "[]" "(::)" ].
Extract Inductive nat => int [ "0" "succ" ].

Extraction "stacks.ml" MyStack.
